import scipy.constants as cst
import scipy.sparse.linalg as linalg
import time
import numpy as np

def solve(P, beta_trial, E_trial=None, neigs=1):
    """
	Solves eigenproblem and returns beta and the transverse E-feilds
	"""
    print('Solving eigenmodes on CPU')
    t = time.time()

    beta_squared, E = linalg.eigs(P, neigs, sigma=beta_trial ** 2, v0 = E_trial)

    Ex, Ey = np.split(E, 2)
    
    Ex, Ey = np.transpose(Ex), np.transpose(Ey)
    print('{} secs later we have the final solution.'.format(time.time() - t))

    return beta_squared ** 0.5, Ex, Ey

def solve_fancy(P, beta_trial, E_trial=None, neigs=1):
    """
    Solves eigenproblem with fancier tools
    """
    from petsc4py import PETSc
    from slepc4py import SLEPc
    

    t = time.time()
    
    # convert eigenproblem into a form that petsc can understand
    fancy_P = PETSc.Mat().createAIJ(size=P.shape,
                                      csr=(P.indptr, P.indices,
                                           P.data))
    
    
    # initalise solver object
    E = SLEPc.EPS(); E.create()
    
    # lets set the solver options
    E.setOperators(fancy_P)
    E.setProblemType(SLEPc.EPS.ProblemType.NHEP)
    E.setDimensions(nev = neigs)
    if E_trial != None:
        E.setInitialSpace(E_trial)

    # now we set up the spectral region to look in 	
    E.setTarget(beta_trial**2) 
    E.setWhichEigenpairs(7) #look for closest in absoult value

    
    
    #beta_trial**2)
    print('Solving eigenmodes using fancy solver')
    E.solve()
    
    #now we can start unpacking	
    nconv = E.getConverged()
    beta = []
    Ex = []
    Ey = []
    vr, wr = fancy_P.getVecs()
    vi, wi = fancy_P.getVecs()
    for i in range(nconv):
        beta.append(E.getEigenpair(i, vr, vi)**0.5)
        #beta.append(E.getEigenvalue(i)**0.5)
        Exr, Eyr = np.split(np.array(vr),2)
        Exi, Eyi = np.split(np.array(vi),2)
        Ex.append(Exr + 1j * Exi) 
        Ey.append(Eyr + 1j * Eyi)
        
        
     
    E.destroy()
    fancy_P.destroy()
    
    return beta, Ex, Ey 



"""
Created on Wed Oct 11 09:36:36 2017

@author: pbm24

fd mode solver based on:
Full-vectorial finite-difference anaylysis of microstructured oprical fibres
Zhaoming Zhu and Thomous G Brown

Please break open the functions and figure out how they work.
"""

from philsol.core import eigen_build

"""
Contains the main maths, lifted from the Zhu paper to assemble Maxwells equations into a linear eigenvalue problem:
- The synatax is as close to verbatim as we could make it
- Returns scipy.sparse matrix which defines the eigen problem for a given refractive index profile. 
- Other operator can be return to help construct other coupled fields (such as E_z, H_x, H_y, H_z)
"""

from philsol import solve

"""
- Handles the eigen solving bit 
- Nothing too complicated here at the moment it just uses standard library scipy sparse libary.
- We hope to add some more fancy eigen solving tricks at a later date (e.g SLEPc )
"""

from philsol import construct

"""
- This function is empty. 
"""



import numpy as np
import scipy.sparse as sps
import time

def eigen_build(k0, n, dx, dy, x_boundary = None, y_boundary = None):
    
    
    # lets find out size of grid and construct some finite difference operators  
    # These can take different forms depending on the user inputed boundarys
    # These operators also need boundaries
    nx, ny, dummy = np.shape(n)
    print('Assembling matrix for {} grid points...\n'.format(nx*ny))
    

    
    # construct finite difference operators single row of FD grid 
    Ux_temp = ( - np.eye(nx, k = 0) + np.eye(nx, k = 1) ) / dx
    
    Uy  = ( - sps.eye(nx*ny, k = 0) + sps.eye(nx*ny, k = nx) ) / dy
  
    if x_boundary == 'periodic':
        Ux_temp[nx-1, 0] = 1. / dx
        
    if y_boundary == 'periodic':
        # This boundary needs a bit more thought although my intuition says its 
        # just a wrap around
        Uy = Uy + sps.eye(nx*ny, k= -(nx-1)*ny) / dy
    
    
    #This statement is kind of confusing but is the equivilent to doing a tensor 
    #contraction. So each row operation is apended to the diagonals of a larger 
    #matrix so we can operate on the whole grid at once.  
    Ux = sps.block_diag([Ux_temp for i in range(ny)], format = 'csr')
    
#%% Now we can construct all the other operators        
 
    Vx = - Ux.transpose()
    Vy = - Uy.transpose()
    I =  sps.eye(nx*ny)    

    
    # We then build relative permitivity tensors  
    epsx = np.empty(nx*ny)
    epsy = np.empty(nx*ny)    
    epszi = np.empty(nx*ny) 
    count = 0 
    for j in range(0,ny):
        for i in range(0, nx):            
            epsx[count] = n[i, j, 0]**2   
            epsy[count] = n[i, j, 1]**2
            epszi[count] = 1. / n[i, j, 2]**2
            count = count + 1 
            
            
    epsx =  sps.spdiags(epsx, 0, nx*ny, nx*ny, format = 'csr')
    epsy =  sps.spdiags(epsy,0, nx*ny, nx*ny, format = 'csr')   
    epszi = sps.spdiags(epszi,0, nx*ny, nx*ny, format = 'csr')
    

    # Now we need to construct the full operator matrices
    t = time.time()     
    Pxx = (- Ux * epszi * Vy * Vx * Uy / k0**2 
             + (k0**2 * I + Ux * epszi * Vx) * (epsx + Vy * Uy / k0**2)    )                   
    
    Pyy = (- Uy * epszi * Vx * Vy * Ux / k0**2 
             + (k0**2 * I + Uy * epszi * Vy) * (epsy + Vx * Ux / k0**2)    )
             
    Pxy = (  Ux * epszi * Vy * (epsy + Vx * Ux / k0**2)
              -  (k0**2 * I + Ux * epszi * Vx) * Vy * Ux / k0**2   )
          
    Pyx = (  Uy * epszi * Vx * (epsx + Vy * Uy / k0**2 ) 
             - (k0**2 * I + Uy * epszi * Vy) * Vx * Uy / k0**2 )  
    
    print('and we are done (after {} secs).'.format(time.time() - t))     
          
    
    # Ok we should be able to do the final assembly now !!!
    P = sps.vstack( [ sps.hstack([Pxx, Pxy]) , sps.hstack([Pyx, Pyy])] ) 

    return P, {'epsx': epsx, 'epsy':epsy, 'epszi': epszi, 
                                       'ux': Ux, 'uy': Uy, 'vx': Vx, 'vy': Vy }




# -*- coding: utf-8 -*-
from distutils.core import setup

packages = \
['philsol']

package_data = \
{'': ['*']}

install_requires = \
['numpy>=1.17,<2.0', 'scipy>=1.3,<2.0']

setup_kwargs = {
    'name': 'philsol',
    'version': '0.22',
    'description': 'A fully vectorial finite difference waveguide mode solver. Based on the algorithm of Zhu and Brown',
    'long_description': "# philsol\n## Modes for the Masses (Massless?)\nFed up with relying on expensive proprietary software for your electromagnetic waveguide research?  philsol might just be the package for you. \nThis is a fully vectorial finite difference waveguide mode solver. In a world where high performance hardware is cheaper than specialist software (including certain commercial packages which may or may not rhyme with my code), philsol throws elegence and sophistication out of the window and replaces it with brute force. \n\nThis is a direct Python implimentaion of the algorithm found in the paper: \n['Full-vectorial finite-difference analysis of microstructured optical fibres', by Zhu and Brown.](https://doi.org/10.1364/OE.10.000853)\n\nWarning: I haven't thoroughly tested so be wary and check the results are sensible...\n\nNew Warning: Original paper by Zhu and Brown is in gaussian not S. I. units. \nThis means that the constructed H fields must be corrected by a factor of $\\mu_0 c$ for calculations in S.I units. \n\n## Installation\n- Clone or download, and install with `python3 setup.py install` (you may need to install python-setuptools)\n- If you are new to python, the important bit is the function in core.py which can be run by itself. \n\n## Examples\n- Commented example projects can be found in the *examples* directory.\n- To run the examples, first install philsol to your Python environment (see above)\n\n## Features\n### Solver\n- Solves vector Maxwell(Helmholtz) equations in 2D for arbitary refractive index profile.\n- Return x and y componants of electric field.\n- philsol can handle anisotropic refractive indices with diagonal tensor.\n- Currently hard coded with conductive boundary.\n- Now includes choice of eigensolver: the default scipy.sparse solver based on ARPACK and Petsc (but you will need a working install of slepc4py and petsc4py)\n- Extra field componants Ez, Hx, Hy, Hz can be calculated from construct module\n- Periodic boundary conditions \n\n### Geometry building\n- The quickest way of importing geometry is with a bitmap image \n- See *examples/example_image.py* for an example in loading .bpm images\n- See *examples/example_build.py* for an example in building geometry using PIL/Pillow\n\n## To do \n- More intelligent geometry aproximation (e.g pixel interpolation on curved boundaries)\n- More boundry condition options Bloch, PML...\n- GPU eigensolving\n\n\n",
    'author': 'Phil Main',
    'author_email': 'philmain28@gmail.com',
    'url': 'https://github.com/philmain28/philsol',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.5,<4.0',
}


setup(**setup_kwargs)



"""
Reconstruct fields from eigenbuild values and solver solutions.

Takes arguments of k0 = 2 pi / wavelength 
and matrices generated by philsol eigenbuild

"""

def extra_feilds(k0, beta, Ex, Ey, matrices):
    
    Hz = 1j * (matrices['uy'] * Ex - matrices['ux'] * Ey) / k0
    Hx = (1j * matrices['vx'] * Hz - k0 * matrices['epsy'] * Ey) / beta
    Hy = (k0 * matrices['epsx'] * Ex - 1j * matrices['vy'] * Hz) / beta
    
    Ez = matrices['epszi'] * (-matrices['vy'] * Hx + matrices['vx'] * Hy) 

    return Ez, Hx, Hy, Hz





"""
Created on Mon Oct  1 14:36:49 2018

Sweet lets wrap up all our functions into a higher level class to save us the 
effort of actualy explicitly passing everything around

The n_build function will need to be manually changed


@author: pbm24
"""
import philsol as ps
import numpy as np

class phil_class:
    def __init__(self, n, k0, x_max = None, y_max = None, dx = None, dy = None):
        self.k0 = k0
        self.n = n
        self.num_x, self.num_y, _ = np.shape(n)
        self.Eigs = 1
        self.E_trial = None
        self.P = None
        
        if x_max == None and y_max == None: 
            if dx != None and dy != None:
               self.dx = dx
               self.dy = dy
               self.x = np.array(range(self.num_x)) * self.dx
               self.y = np.array(range(self.num_y)) * self.dy
            else: 
               raise Exception('Gonna need some dimensions yo!')
               
        if dx == None and dy == None: 
            if x_max != None and y_max != None:
               self.dx = x_max / float(self.num_x - 1)
               self.dy = y_max / float(self.num_y - 1)
            else: 
               raise Exception('Gonna need some dimensions yo!')


    def build_stuff(self, x_bound = None, y_bound = None, matrices = None):
        
        
        if matrices == None:
            self.P, _ = ps.core.eigen_build(    self.k0, 
                                                self.n, 
                                                self.dx, 
                                                self.dy, 
                                                x_boundary = x_bound, 
                                                y_boundary = y_bound      )
        else:
            self.P, self.mats  =  ps.core.eigen_build(    self.k0, 
                                                          self.n, 
                                                          self.dx, 
                                                          self.dy, 
                                                          x_boundary = x_bound, 
                                                          y_boundary = y_bound)
    
    def solve_stuff(self, neigs, beta_trial, extra_fields = None):
        '''
        Function to pass object to solver with the option of constructing the 
        ez and h feilds
        '''
        
        # some insults if everything isn't set up
        if self.P == None: 
            print("Build your eigenproblem you idiot")
            
        
        self.Eigs = neigs
        self.beta, self.Ex, self.Ey = ps.solve.solve(   self.P, 
                                                        beta_trial, 
                                                        E_trial = self.E_trial, 
                                                        neigs = self.Eigs    )

        if extra_fields == True:
            self.E = np.empty(self.neigs, 2 * self.num_x * self.num_y, 3)
            self.H = np.empty_like(self.E)
            for i in self.neigs:
                self.E[i, :, 0] = self.Ex[i,:]
                self.E[i, :, 1] = self.Ey[i,:]
                ez, hx, hy, hz = ps.construct()
                self.E[i, :, 2] = ez
                self.H[i, :, 0] = hx
                self.H[i, :, 1] = hy        
                self.H[i, :, 2] = hz        

        
        
